//NOTES: WAL
//http://www.sqlite.org/wal.html
//for better perfomance, but more space to consume...:
//db.exec("PRAGMA journal_mode = WAL");
//if have more even bigger space:
//db.exec("PRAGMA wal_autocheckpoint=N");

var module_name='q_sqlite3';
var debug=1;
var logger=console;
//const sqlite3 = require('sqlite3').verbose();
const sqlite3 = require('sqlite3');
const Q=require('q');

var db;

function qstr(s){ if(s==null) return "''"; return "'"+(""+s).replace(/'/g,"''")+"'"; }

function qstr_arr(a){ var rt_a=[]; for(var k in a){ rt_a.push(qstr(a[k])); } return rt_a.join(',') }

module.exports = db_name=>{

	if(db_name) db=new sqlite3.Database(db_name);

	function select_Promise(sql){
		var dfr=Q.defer();
		db.all(sql,function(err,rows){
			if(err) dfr.reject(err);
			else dfr.resolve({STS:'OK',rows});
		});
		return dfr.promise;
	}

	var exec_Promise=function(sql,max_try_for_busy){
		if(!sql)return Q({STS:'KO',errmsg:'exec_Promise meets a empty .sql?'});
		sql=sql.trim();

		if(sql.substr(0,6).toUpperCase()=='SELECT') return select_Promise(sql);

		if('undefined'==typeof max_try_for_busy) max_try_for_busy=7;

		var dfr=Q.defer();
		db.serialize(function(){
			db.run(sql,function(err,rst){
				if(err){
					var {code}=err;
					if('SQLITE_BUSY'==code){
						if(max_try_for_busy>0){
							Q.delay(Math.random()*50+100).done(()=>{
								if(debug>0) logger.log('SQLITE_BUSY_RETRY/'+max_try_for_busy,sql);
								exec_Promise(sql,max_try_for_busy-1)
									.fail(err=>{
										if(debug>1) logger.log('SQLITE_BUSY_C',max_try_for_busy,sql);
										dfr.reject(err)
									})
									.done(rst=>{ dfr.resolve(rst) });
							});
						}else{
							if(debug>1) logger.log('SQLITE_BUSY_B',max_try_for_busy,sql);
							dfr.reject(err);
						}
					}else{
						if(debug>0) logger.log('SQLITE_ERR_A',code,max_try_for_busy,sql);
						dfr.reject(err);
					}
				}else{
					var {lastID,changes}=this;//@ref https://github.com/mapbox/node-sqlite3/wiki/API
					dfr.resolve({STS:'OK',sql,lastID,changes,af:changes});
				}
			});
		});
		return dfr.promise;
	};
	//"_Promise" suffix is deprecated, now _q era !
	var exec_q = exec_Promise;
	var select_q = select_Promise;

	//our UPSERT algorithm FYR
	var FindAndUpsert_Promise =function(table,toUpdate,toFind,insert_first){
		var s_kv="",a_kv=[],
			s_w="",a_w=[],
			s_v="",a_v=[],
			s_k="",a_k=[];
		for(var k in toFind){
			var v=toFind[k];
			a_w.push(""+k+"="+qstr(v));
		}
		var where="WHERE "+a_w.join(" AND ");
		for(var k in toUpdate){
			var v=toUpdate[k];
			a_v.push(qstr(v)+" AS "+k);
			a_k.push(qstr(k));
			a_kv.push(""+k+"="+qstr(v));
		}
		s_k=a_k.join(",");
		s_v=a_v.join(",");
		s_kv=a_kv.join(",");

		var tmp_table='TMP_'+(new Date()).getTime();

		var sql_1 = `INSERT INTO ${table} (${s_k})
		SELECT * FROM (SELECT ${s_v}) AS ${tmp_table}
		WHERE NOT EXISTS (SELECT 'Y' FROM ${table} ${where} LIMIT 1)`;

		var sql_2=`UPDATE ${table} SET ${s_kv} ${where}`;

		var lastID=-1;
		var af=-1;

		if(insert_first){
			//try insert first
			return exec_Promise(sql_1)
				.then(rst=>{
					lastID=rst.lastID;
					return exec_Promise(sql_2)
						.then(rst=>{
							af=rst.af;
							return Q({STS:af>0?'OK':'KO',lastID,af})
						})
				})
				.fail(err=>{
					if(debug>0)
						logger.log('DEBUG FindAndUpsert_Promise.err=',err,sql_1,sql_2);
					return Q.reject(err)
				})
		}else{
			//try update first (default)
			return exec_Promise(sql_2)
				.then(rst=>{
					af=rst.af;
					if(af>0){
						return Q({STS:"OK",lastID,af})
					}else{
						return exec_Promise(sql_1)
							.then(rst=>{
								lastID=rst.lastID;
								return Q({STS:lastID>0?'OK':'KO',lastID,af})
							})
					}
				})
				.fail(err=>{
					if(debug>0)
						logger.log('DEBUG FindAndUpsert_Promise.err=',err,sql_1,sql_2);
					return Q.reject(err)
				})
		}
	};
	var FindAndUpsert_q = FindAndUpsert_Promise;

	return {
		db,
		qstr,qstr_arr,//expose two helper func
		exec_Promise,exec_q,
		select_Promise,select_q,
		FindAndUpsert_Promise,FindAndUpsert_q,
		setDebugLevel(d){
			if(d>0)logger.log(module_name+'.setDebugLevel=',d);
			debug=d;
		}
	}
}
